/*----------------------------------------------------------------------
  File    : chull.c
  Contents: determine a convex hull in 3-space
  Author  : Christian Borgelt
  History : 2000.10.19 file created
            2000.10.21 function ch_start() completed
            2000.10.22 function ch_next()  completed
            2000.10.23 merging coplanar facets added
            2000.11.05 functions for sorted points algorithm added
            2000.11.07 function ch_sorted() completed
            2000.12.09 drawing as a solid body added
            2000.12.10 sorted points algorithm debugged
            2001.06.16 bug in drawing function removed
            2013.08.09 adapted to higher compiler warning level
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "lists.h"
#include "chull.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define EPSILON  1e-14          /* to avoid roundoff error problems */

/*----------------------------------------------------------------------
  Conflict List Functions
----------------------------------------------------------------------*/

static CFLE* add_confl (POINT3 *point, FACET *facet)
{                               /* --- add a conflict to a c. hull */
  CFLE *cfle;                   /* created conflict list element */

  cfle = (CFLE*)malloc(sizeof(CFLE));
  if (!cfle) return NULL;       /* create a conflict list element */
  cfle->point   = point;        /* and store point and facet */
  cfle->facet   = facet;
  if (point->confls) point->confls->pt_pred = cfle;
  cfle->pt_succ = point->confls;/* add the new element at the head */
  cfle->pt_pred = NULL;         /* of the point's conflict list */
  point->confls = cfle;
  if (facet->confls) facet->confls->fc_pred = cfle;
  cfle->fc_succ = facet->confls;/* add the new element at the head */
  cfle->fc_pred = NULL;         /* of the facet's conflict list */
  facet->confls = cfle;
  return cfle;                  /* return the created list element */
}  /* add_confl() */

/*----------------------------------------------------------------------
  Point Functions
----------------------------------------------------------------------*/

static POINT3* add_point (CHULL *chull, double x, double y, double z)
{                               /* --- add a point to a convex hull */
  POINT3 *pt;                   /* created point */

  assert(chull);                /* check the function arguments */
  pt = (POINT3*)malloc(sizeof(POINT3));
  if (!pt) return NULL;         /* create a point structure */
  pt->x = x; pt->y = y; pt->z = z;
  pt->refcnt = 0;               /* store the point's coordinates */
  pt->confls = NULL;            /* and initialize remaining fields */

  if (chull->points) chull->points->pred = pt;
  pt->succ = chull->points;     /* add the new point to the list */
  pt->pred = NULL;              /* and increment the point counter */
  chull->points = pt; chull->ptcnt++;
  return pt;                    /* return the created point */
}  /* add_point() */

/*--------------------------------------------------------------------*/

static void del_point (CHULL *chull, POINT3 *pt)
{                               /* --- delete a point from a c. hull */
  assert(chull && pt);          /* check the function arguments */
  if (--pt->refcnt > 0)         /* if there are still references */
    return;                     /* to this point, abort the function */
  assert(!pt->confls);          /* check for an empty conflict list */
  if (pt->succ) pt->succ->pred = pt->pred;
  if (pt->pred) pt->pred->succ = pt->succ;
  else          chull->points  = pt->succ;
  chull->ptcnt--;               /* remove the point from the list */
  free(pt);                     /* and delete the point structure */
}  /* del_point() */

/*--------------------------------------------------------------------*/

static int cmp_points (const POINT3 *a, const POINT3 *b, void *data)
{                               /* --- compare two points */
  if (a->z > b->z) return +1;
  if (a->z < b->z) return -1;
  if (a->x > b->x) return +1;
  if (a->x < b->x) return -1;
  if (a->y > b->y) return +1;
  if (a->y < b->y) return -1;   /* return the sign of the */
  return 0;                     /* 'difference' between the points */
}  /* cmp_points() */

/*----------------------------------------------------------------------
  Edge Functions
----------------------------------------------------------------------*/

static EDGE* add_edge (CHULL *chull, POINT3 *p0, POINT3 *p1)
{                               /* --- add an edge to a convex hull */
  EDGE *edge;                   /* created edge */

  assert(chull && p0 && p1);    /* check the function arguments */
  edge = (EDGE*)malloc(sizeof(EDGE));
  if (!edge) return NULL;       /* allocate memory for an edge */
  edge->dx = p1->x -p0->x;      /* compute the direction vector */
  edge->dy = p1->y -p0->y;      /* of the edge (from p0 to p1) */
  edge->dz = p1->z -p0->z;
  edge->points[0] = p0; p0->refcnt++;  /* store the two points and */
  edge->points[1] = p1; p1->refcnt++;  /* increment the ref. counters */
  edge->refcnt = 0;             /* init. the edge's reference counter */

  if (chull->edges) chull->edges->pred = edge;
  else              chull->last        = edge;
  edge->succ = chull->edges;    /* add the new edge to the edge list */
  edge->pred = NULL;            /* and increment the edge counter */
  chull->edges = edge; chull->edgecnt++;
  return edge;                  /* return the created edge */
}  /* add_edge() */

/*--------------------------------------------------------------------*/

static void del_edge (CHULL *chull, EDGE *edge)
{                               /* --- delete an edge from a c. hull */
  assert(chull && edge);        /* check the function arguments */
  if (--edge->refcnt > 0)       /* if there are still references */
    return;                     /* to this edge, abort the function */
  if (edge->succ) edge->succ->pred = edge->pred;
  else            chull->last      = edge->pred;
  if (edge->pred) edge->pred->succ = edge->succ;
  else            chull->edges     = edge->succ;
  chull->edgecnt--;             /* remove the edge from the list */
  del_point(chull, edge->points[0]);
  del_point(chull, edge->points[1]);
  free(edge);                   /* try to delete the end points */
}  /* del_edge() */             /* and delete the edge structure */

/*--------------------------------------------------------------------*/

static void reverse (EDGE *edge)
{                               /* --- reverse the dir. of an edge */
  POINT3 *point;                /* buffer for end point */
  FACET  *facet;                /* buffer for incident facet */
  int    pos;                   /* buffer for edge vector position */

  assert(edge);                 /* check the function argument */
  edge->dx = -edge->dx;         /* reverse the direction vector */
  edge->dy = -edge->dy;
  edge->dz = -edge->dz;
  point           = edge->points[0];
  edge->points[0] = edge->points[1];
  edge->points[1] = point;      /* exchange the two end points */
  facet           = edge->facets[0];
  edge->facets[0] = edge->facets[1];
  edge->facets[1] = facet;      /* exchange the two incident facets */
  pos             = edge->pos[0];
  edge->pos[0]    = edge->pos[1];
  edge->pos[1]    = pos;        /* exchange the edge vector positions */
}  /* reverse() */

/*----------------------------------------------------------------------
  Facet Functions
----------------------------------------------------------------------*/

static FACET* add_facet (CHULL *chull, EDGE *e0, EDGE *e1, EDGE *e2)
{                               /* --- add a facet to a convex hull */
  FACET *facet;                 /* created facet */

  assert(chull && e0 && e1 && e2);  /* check the function arguments */
  assert((e0->refcnt < 2)       /* check the number of times */
  &&     (e1->refcnt < 2)       /* the edges have already been used */
  &&     (e2->refcnt < 2));     /* (number of refs. from facets) */
  assert((e0->points[1-e0->refcnt] == e1->points[e1->refcnt])
  &&     (e1->points[1-e1->refcnt] == e2->points[e2->refcnt])
  &&     (e2->points[1-e2->refcnt] == e0->points[e0->refcnt]));
                                /* check for a closed edge list */
  facet = (FACET*)malloc(sizeof(FACET));
  if (!facet) return NULL;      /* allocate memory for a facet */
  facet->edges = (EDGE**)malloc(4 *sizeof(EDGE*));
  if (!facet->edges) { free(facet); return NULL; }
  facet->nx = e0->dy *e1->dz -e0->dz *e1->dy;  /* compute a */
  facet->ny = e0->dz *e1->dx -e0->dx *e1->dz;  /* normal vector */
  facet->nz = e0->dx *e1->dy -e0->dy *e1->dx;  /* of the facet */
  if (e0->refcnt != e1->refcnt) {
    facet->nx = -facet->nx;     /* if one of the first two edges */
    facet->ny = -facet->ny;     /* is to be reversed for a traversal, */
    facet->nz = -facet->nz;     /* the normal vector must be inverted */
  }                             /* to make it point to the outside */
  e0->facets[e0->refcnt] = facet;  /* set the facet links */
  e1->facets[e1->refcnt] = facet;  /* from the edges and */
  e2->facets[e2->refcnt] = facet;  /* then store the edges */
  facet->edges[0] = e0; e0->pos[e0->refcnt++] = 0;
  facet->edges[1] = e1; e1->pos[e1->refcnt++] = 1;
  facet->edges[2] = e2; e2->pos[e2->refcnt++] = 2;
  facet->edgecnt  = 3;           /* note the edge vector positions, */
  facet->edgevsz  = 4;           /* set the edge counter and the */
  facet->confls   = NULL;        /* current edge vector size, and */
  facet->visible  = 0;           /* initialize the remaining fields */

  if (chull->facets) chull->facets->pred = facet;
  facet->succ = chull->facets;  /* add the new facet to the list */
  facet->pred = NULL;           /* and increment the facet counter */
  chull->facets = facet; chull->facetcnt++;
  chull->created++;             /* count the created facets */
  return facet;                 /* return the created facet */
}  /* add_facet() */

/*--------------------------------------------------------------------*/

static void del_facet (CHULL *chull, FACET *facet)
{                               /* --- delete a facet from a c. hull */
  int  i;                       /* loop variable */
  CFLE *cfle, *tmp;             /* to traverse the conflict list */
  EDGE *edge;                   /* to traverse the incident facets */

  assert(chull && facet);       /* check the function arguments */
  if (facet->succ) facet->succ->pred = facet->pred;
  if (facet->pred) facet->pred->succ = facet->succ;
  else             chull->facets     = facet->succ;
  chull->facetcnt--;            /* remove the facet from the list */

  for (cfle = facet->confls; cfle; cfle = cfle->fc_succ) {
    if (cfle->pt_succ) cfle->pt_succ->pt_pred = cfle->pt_pred;
    if (cfle->pt_pred) cfle->pt_pred->pt_succ = cfle->pt_succ;
    else               cfle->point->confls    = cfle->pt_succ;
  }                             /* remove conflicts from point lists */
  for (cfle = facet->confls; cfle;){ /* and delete the conflict list */
    tmp = cfle; cfle = cfle->fc_succ; free(tmp); }

  for (i = facet->edgecnt; --i >= 0; ) {
    edge = facet->edges[i];     /* traverse the incident edges */
    if ((edge->facets[0] == facet)   /* if an edge still refers to */
    ||  (edge->facets[1] == facet))  /* the facet (no horizon edge), */
      del_edge(chull, edge);         /* try to delete that edge */
  }
  free(facet->edges);           /* delete the edge vector */
  free(facet);                  /* and the facet structure */
}  /* del_facet() */

/*--------------------------------------------------------------------*/

static void top_facet (CHULL *chull, FACET *facet)
{                               /* --- move facet to the top (head) */
  assert(chull && facet);       /* check the function arguments */
  if (!facet->pred) return;     /* if it is at the top, abort */
  facet->pred->succ = facet->succ;
  if (facet->succ) facet->succ->pred = facet->pred;
  facet->succ = chull->facets;  /* move the facet to the head */
  facet->pred = NULL;           /* (or top) of the facet list */
  if (chull->facets) chull->facets->pred = facet;
  chull->facets = facet;        /* set the new first element */
}  /* top_facet() */

/*--------------------------------------------------------------------*/

static void red_facet (CHULL *chull, FACET *facet, EDGE *edge)
{                               /* --- reduce a facet of a c. hull */
  int  i, k;                    /* vector indices */
  EDGE *last;                   /* last edge in the edge vector */

  assert(chull && facet && edge);  /* check the function arguments */
  assert(((edge->refcnt > 0) && (edge->facets[0] == facet))
      || ((edge->refcnt > 1) && (edge->facets[1] == facet)));
  last = facet->edges[--facet->edgecnt];
  if (last != edge) {           /* if the edge is not the last one */
    i = (edge->facets[0] == facet) ? 0 : 1;
    k = (last->facets[0] == facet) ? 0 : 1;
    facet->edges[i = edge->pos[i]] = last;
    last->pos[k] = i;           /* shift the last edge to */
  }                             /* where the given edge is */
  del_edge(chull, edge);        /* and delete the given edge */
}  /* red_facet() */

/*--------------------------------------------------------------------*/

static int ext_collin (CHULL *chull, FACET *facet, EDGE *edge)
{                               /* --- extend a collinear edge */
                                /*     of a facet (if there is one) */
  int    i;                     /* loop variable, vector index */
  EDGE   **e;                   /* to traverse the edge vector */
  FACET  *incid;                /* other incident facet */
  POINT3 **a, **b, *t;          /* to exchange the points */

  assert(facet && edge          /* check the function arguments */
      && (edge->refcnt == 1));  /* and the edge's reference counter */
  incid = edge->facets[0];      /* get the other incident facet */
  for (e = facet->edges +(i = facet->edgecnt); --i >= 0; ) {
    if ((*--e)->refcnt < 2)     /* traverse the edges of the facet */
      continue;                 /* except those without another facet */
    if ((*e)->facets[((*e)->facets[0] == facet) ? 1 : 0] != incid)
      continue;                 /* skip edges incident to diff. facet */
    a = (*e)->points;           /* get the point vectors */
    b = edge->points;           /* of the two collinear edges */
    if ((a[0] != b[0])          /* and determine the points */
    &&  (a[0] != b[1])) a++;    /* that have to be exchanged */
    if  (a[0] == b[0])  b++;    /* in order to extend the edge *e, */
    t = *a; *a = *b; *b = t;    /* then exchange these points */
    /* At this point 'edge' refers to the same point twice, */
    /* therefore deleting this edge also deletes the point. */
    red_facet(chull, incid, edge);
    return 0;                   /* remove the edge from the */
  }                             /* incident facet and delete it */
  return -1;                    /* return 'no collinear edge found' */
}  /* ext_collin() */

/*--------------------------------------------------------------------*/

static FACET* ext_facet (CHULL *chull, FACET *facet,
                         EDGE *out, EDGE *in1, EDGE *in2)
{                               /* --- extend a facet of a c. hull */
  int  size;                    /* current size of the edge vector */
  EDGE **tmp;                   /* buffer for a new edge vector */

  assert(facet && out && in1 && in2); /* check the function arguments */
  assert(out->facets[0] == facet);
  assert((out->refcnt <= 2)     /* check the number of times */
  &&     (in1->refcnt <  2)     /* the edges have already been used */
  &&     (in2->refcnt <  2));   /* (number of refs. from facets) */
  assert((chull->facetcnt == 1) /* check the edges' end points */
  ||     ((out->points[0]             == in1->points[in1->refcnt])
  &&      (in1->points[1-in1->refcnt] == in2->points[in2->refcnt])
  &&      (in2->points[1-in2->refcnt] == out->points[1])));

  red_facet(chull, facet, out); /* delete the facet to be removed */
  if (((in1->refcnt > 0) && (in1->facets[0] == facet))
  ||  ((in1->refcnt > 1) && (in1->facets[1] == facet))) {
    red_facet(chull, facet, in1);
    in1 = NULL;                 /* if the first edge is incident */
  }                             /* to the facet, delete it */
  if (((in2->refcnt > 0) && (in2->facets[0] == facet))
  ||  ((in2->refcnt > 1) && (in2->facets[1] == facet))) {
    red_facet(chull, facet, in2);
    in2 = NULL;                 /* if the second edge is incident */
  }                             /* to the facet, delete it */
  size = facet->edgevsz;        /* get the size of the edge vector */
  if (facet->edgecnt+2 > size){ /* and check whether it is full */
    size += (size > 4) ? size >> 1 : 4;
    tmp   = (EDGE**)realloc(facet->edges, (size_t)size *sizeof(EDGE*));
    if (!tmp) return NULL;      /* enlarge the edge vector */
    facet->edges = tmp; facet->edgevsz = size;
  }                             /* set the new vector and its size */
  if (  in1                     /* if to add edge in1 to the facet */
  &&  ((in1->refcnt < 1)        /* and there is no collinear edge */
  ||   (ext_collin(chull, facet, in1) != 0))) {
    in1->facets [in1->refcnt  ]    = facet;
    in1->pos    [in1->refcnt++]    = facet->edgecnt;
    facet->edges[facet->edgecnt++] = in1;
  }                             /* add edge and update facet */
  if (  in2                     /* if to add edge in2 to the facet */
  &&  ((in2->refcnt < 1)        /* and there is no collinear edge */
  ||   (ext_collin(chull, facet, in2) != 0))) {
    in2->facets [in2->refcnt  ]    = facet;
    in2->pos    [in2->refcnt++]    = facet->edgecnt;
    facet->edges[facet->edgecnt++] = in2;
  }                             /* add edge and update facet */
  return facet;                 /* return the extended facet */
}  /* ext_facet() */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void cleanup (CHULL *chull, int delpts)
{                               /* --- clean up (after an error) */
  POINT3 *p, *tp;               /* to traverse the list of points */
  EDGE   *e, *te;               /* to traverse the list of edges */
  FACET  *f, *tf;               /* to traverse the list of facets */
  CFLE   *c, *tc;               /* to traverse the conflict list */

  assert(chull);                /* check the function argument */
  for (f = chull->facets; f;) { /* traverse the list of facets */
    for (c = f->confls; c;) {   /* delete the conflict list */
      tc = c; c = c->fc_succ; free(tc); }
    free(f->edges);             /* delete the edge vector */
    tf = f; f = f->succ; free(tf);
  }                             /* delete the facet structure */
  chull->facets = NULL; chull->facetcnt = 0;
  for (e = chull->edges; e; ) { /* delete the list of edges */
    te = e; e = e->succ; free(te); }
  chull->edges = NULL; chull->edgecnt = 0;
  if (delpts) {                 /* if to delete the points */
    for (p = chull->points; p; ) {
      tp = p; p = p->succ; free(tp); }
    chull->points = NULL; chull->ptcnt = 0; }
  else {                        /* if to keep the points */
    for (p = chull->points; p; p = p->succ) {
      p->refcnt = 0;            /* traverse the point list and */
      p->confls = NULL;         /* clear the reference counter */
    }                           /* and the conflict list pointer */
  }
  chull->created = 0;           /* clear the created facets counter */
  chull->next = NULL;           /* clear the next unprocessed point */
}  /* cleanup() */

/*--------------------------------------------------------------------*/

static int visib2d (FACET *facet, EDGE *edge, POINT3 *pt)
{                               /* --- check visibility of an edge */
                                /*     in a 2-dimensional convex hull */
  POINT3 *base;                 /* base point of the edge */
  double x, y, z;               /* vector coordinates */

  assert(facet && edge && pt);  /* check the function arguments */
  base = edge->points[0];       /* get the edge's base point */
  x    = pt->x -base->x;        /* compute the vector product of the */
  y    = pt->y -base->y;        /* edge and a vector to the point and */
  z    = pt->z -base->z;        /* project this onto the facet normal */
  return (facet->nx *(y *edge->dz -z *edge->dy)
        + facet->ny *(z *edge->dx -x *edge->dz)
        + facet->nz *(x *edge->dy -y *edge->dx) >= 0);
}  /* visib2d() */              /* return whether proj. is not neg. */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

CHULL* ch_create (void)
{                               /* --- create a convex hull structure */
  return (CHULL*)calloc(1, sizeof(CHULL));
}  /* ch_create() */

/*--------------------------------------------------------------------*/

void ch_delete (CHULL *chull)
{                               /* --- delete a convex hull structure */
  cleanup(chull, 1);            /* clean up the convex hull structure */
  free(chull);                  /* and delete the structure itself */
}  /* ch_delete() */

/*--------------------------------------------------------------------*/

void ch_rand (int type, RANDFN randfn, double *x, double *y, double *z)
{                               /* --- generate a random 3D point */
  double a, b, c;               /* random numbers */
  double t = 0;                 /* temporary buffer */

  assert(randfn && x && y && z);/* check the function arguments */
  do {                          /* point generation loop */
    a = 2 *randfn() -1;         /* generate three random coordinates */
    b = 2 *randfn() -1;         /* in the range [-1,1] */
    c = 2 *randfn() -1;
    if ((type != CH_SPHERE)     /* if a point not in a sphere or on */
    &&  (type != CH_SURFACE))   /* the surface of a sphere is req., */
      break;                    /* abort the loop, otherwise */
    t = sqrt(a*a +b*b +c*c);    /* compute distance from origin */
  } while (t > 1);              /* while point is not inside sphere */
  if (type == CH_SURFACE) {     /* normalize distance to get a point */
    *x = a/t; *y = b/t; *z = c/t; }    /* on the surface of a sphere */
  else {                        /* for points inside a sphere and */
    *x = a;   *y = b;   *z = c;}/* inside a cube, simply copy the */
}  /* ch_rand() */              /* coordinates */

/*--------------------------------------------------------------------*/

int ch_point (CHULL *chull, double x, double y, double z)
{                               /* --- add a 3D point */
  return (add_point(chull, x, y, z)) ? 0 : -1;
}  /* ch_point() */

/*--------------------------------------------------------------------*/

int ch_start (CHULL *chull, int alg)
{                               /* --- start building a convex hull */
  POINT3 *a, *b, *c, *d;        /* corners of the initial tetrahedron */
  EDGE   *ab, *bc, *ca,         /* edges   of the initial tetrahedron */
         *ad, *db, *dc;
  FACET  *abc, *facet;          /* facets  of the initial tetrahedron */
  EDGE   *edge;                 /* to traverse the list of edges */
  double x, y, z, k;            /* vector coordinates, buffer */
  int    pos;                   /* buffer for position in edge vector */

  assert(chull);                /* check the function arguments */
  if (chull->edges)             /* check whether building a convex */
    return -2;                  /* hull has already been started */
  chull->init = chull->ptcnt;   /* note the initial number of points */
  chull->alg  = alg;            /* note the algorithm to use and */
  if (alg == CH_SORTED)         /* if to use the sorted points alg. */
    chull->points = l_sort(chull->points, (CMPFN*)cmp_points, NULL);

  a = chull->points;            /* get the first point in the list */
  if (!a) return 1;             /* if there is none, abort */
  if (alg == CH_SORTED) {       /* if to use the sorted points alg. */
    for (b = a, c = a->succ; c; c = b->succ) {
      if ((fabs(c->x -b->x) <= EPSILON)
      &&  (fabs(c->y -b->y) <= EPSILON)
      &&  (fabs(c->z -b->z) <= EPSILON))
        del_point(chull, c);    /* traverse the sorted list of points */
      else b = c;               /* and delete all duplicate points */
    }                           /* (in the sorted points algorithm */
  }                             /* duplicates cannot be handled) */

  /* --- find two points that are different --- */
  while (1) {                   /* search loop */
    b = a->succ;                /* get the next (second) point */
    if (!b) return 1;           /* if there is none, abort */
    if ((fabs(b->x -a->x) > EPSILON)
    ||  (fabs(b->y -a->y) > EPSILON)
    ||  (fabs(b->z -a->z) > EPSILON))
      break;                    /* if the point is different, */
    del_point(chull, b);        /* abort the loop, otherwise */
  }                             /* delete the point */
  chull->dim++;                 /* convex hull has one dimension */

  /* --- find three points that are not collinear --- */
  while (1) {                   /* search loop */
    c = b->succ;                /* get the next (third) point */
    if (!c) break;              /* if there is none, abort the loop */
    x = b->x -a->x;             /* determine the direction vector */
    y = b->y -a->y;             /* from a to b and compare it to */
    z = b->z -a->z;             /* the direction vector from a to c */
    if      (fabs(x) > EPSILON) k = (c->x -a->x) /x;
    else if (fabs(y) > EPSILON) k = (c->y -a->y) /y;
    else                        k = (c->z -a->z) /z;
    if ((c->x -a->x != k *x)    /* if the three points a, b, c */
    ||  (c->y -a->y != k *y)    /* are not collinear (do not */
    ||  (c->z -a->z != k *z))   /* lie on a straight line), */
      break;                    /* abort the search loop */
    if      (k < 0) { del_point(chull, a); a = b; b = c; }
    else if (k > 1) { del_point(chull, b);        b = c; }
    else            { del_point(chull, c);               }
  }                             /* delete the point in the middle */

  /* --- create first facet of tetrahedron --- */
  ab = add_edge(chull, a, b);   /* create an edge from a to b */
  if (!ab)  { cleanup(chull, 0); return -1; }
  if (!c) return 1;             /* if there are only 2 points, abort */
  bc = add_edge(chull, b, c);   /* create an edge from b to c */
  if (!bc)  { cleanup(chull, 0); return -1; }
  ca = add_edge(chull, c, a);   /* create an edge from c to a */
  if (!ca)  { cleanup(chull, 0); return -1; }
  abc = add_facet(chull, ab, bc, ca);     /* create facet abc */
  if (!abc) { cleanup(chull, 0); return -1; }

  chull->dim++;                 /* convex hull has two dimensions */
  if (!c->succ) return 1;       /* if there are only 3 points, abort */
  if (alg == CH_SORTED) {       /* if to use the sorted points alg., */
    chull->next = c->succ; return 0; }         /* abort the function */

  /* --- find four points that are not coplanar --- */
  d = c;                        /* start from the third point */
  do {                          /* search loop */
    d = d->succ;                /* get the next point */
    if (!d) {                   /* if all points are coplanar, */
      cleanup(chull, 0);        /* switch to the sorted points alg. */
      return ch_start(chull, CH_SORTED);
    }
    x = d->x -a->x;             /* the four points are not coplanar */
    y = d->y -a->y;             /* if the vector from a to d is not */
    z = d->z -a->z;             /* orthogonal to the facet normal */
    k = x *abc->nx +y *abc->ny +z *abc->nz;
  } while (fabs(k) <= EPSILON); /* while point is in the plane */

  if (d->succ) d->succ->pred = d->pred;
  d->pred->succ = d->succ;      /* move the point found */
  d->succ = c->succ;            /* behind the third point */
  d->pred = c; c->succ = d;     /* in the point list */
  if (d->succ) d->succ->pred = d;

  /* --- create remaining edges of tetrahedron --- */
  ad = add_edge(chull, a, d);   /* create an edge from d to a */
  if (!ad) { cleanup(chull, 0); return -1; }
  db = add_edge(chull, d, b);   /* create an edge from d to b */
  if (!db) { cleanup(chull, 0); return -1; }
  dc = add_edge(chull, d, c);   /* create an edge from d to c */
  if (!dc) { cleanup(chull, 0); return -1; }

  /* --- create remaining facets of tetrahedron --- */
  if (!add_facet(chull, ab, ad, db)) {
    cleanup(chull, 0); return -1; }  /* create facet abd */
  if (!add_facet(chull, bc, db, dc)) {
    cleanup(chull, 0); return -1; }  /* create facet bcd */
  if (!add_facet(chull, ca, dc, ad)) {
    cleanup(chull, 0); return -1; }  /* create facet acd */

  /* --- orient normal vectors and edges --- */
  if (k > 0) {                  /* check the side of the point */
    for (facet = chull->facets; facet; facet = facet->succ) {
      facet->nx = -facet->nx;   /* traverse the created faces */
      facet->ny = -facet->ny;   /* and invert the normal vectors, */
      facet->nz = -facet->nz;   /* since the current ones point */
    }                           /* to the inside of the tetrahedron */
    for (edge = chull->edges; edge; edge = edge->succ) {
      facet           = edge->facets[0];
      edge->facets[0] = edge->facets[1];
      edge->facets[1] = facet;
      pos             = edge->pos[0];
      edge->pos[0]    = edge->pos[1];
      edge->pos[1]    = pos;    /* traverse the created edges */
    }                           /* and exchange the incident facets */
  }                             /* (for counterclockwise direction) */
  chull->dim++;                 /* convex hull has three dimensions */

  /* --- build conflict lists --- */
  for (b = d->succ; b; b = b->succ) {
    for (facet = chull->facets; facet; facet = facet->succ) {
      a = facet->edges[0]->points[0];
      x = b->x -a->x;           /* traverse all combinations of */
      y = b->y -a->y;           /* points and facets and check */
      z = b->z -a->z;           /* whether there is a conflict */
      if (x *facet->nx +y *facet->ny +z *facet->nz <= EPSILON)
        continue;               /* if there is no conflict, continue */
      if (!add_confl(b, facet)) { cleanup(chull, 0); return -1; }
    }                           /* otherwise create a conflict */
  }                             /* list element from point and facet */

  chull->next = d->succ;        /* note the next unprocessed point */
  return (chull->next) ? 0 : 1; /* return whether convex hull */
}  /* ch_start() */             /* is complete (1) or not (0) */

/*--------------------------------------------------------------------*/

static int ch_2dim (CHULL *chull)
{                               /* --- process next point in 2 dim. */
  int    i;                     /* edge index */
  POINT3 *curr, *base;          /* current point, base of an edge */
  EDGE   **p;                   /* to reallocate the edge vector */
  EDGE   *edge;                 /* to traverse the list of edges */
  FACET  *facet;                /* the only facet of the c. hull */
  EDGE   *r2c, *c2l;            /* edges to/from the horizon points */
  double x, y, z, k;            /* vector coordinates, buffer */

  assert(chull && chull->next); /* check the function argument */
  assert(chull->dim == 2);      /* check the dimension of the c. hull */

  /* --- check whether next point is still in the plane --- */
  curr  = chull->next;          /* get the next unprocessed point */
  facet = chull->facets;        /* and the only facet */
  base  = chull->points;        /* and a base point of the facet */
  x     = curr->x -base->x;     /* compute a difference vector from */
  y     = curr->y -base->y;     /* the current point to the base and */
  z     = curr->z -base->z;     /* project it to the facet normal */
  k     = x *facet->nx +y *facet->ny +z *facet->nz;
  if (fabs(k) > EPSILON) {      /* if the c. hull is no longer planar */
    if (chull->edgecnt > facet->edgevsz) {
      p = (EDGE**)realloc(facet->edges,
                         (size_t)chull->edgecnt *sizeof(EDGE*));
      if (!p) { cleanup(chull, 0); return -1; }
      facet->edgevsz = chull->edgecnt;
      facet->edges   = p;       /* if the facet's edge vector */
    }                           /* is too small, enlarge it */
    p = facet->edges; i = 0;    /* traverse all edges of the facet */
    for (edge = chull->edges; edge; edge = edge->succ) {
      *p++ = edge;              /* store the edge in the vector */
      edge->facets[0] = facet;  /* note a reference to the facet */
      edge->pos   [0] = i++;    /* set the edge's vector position */
      edge->refcnt    = 1;      /* and its reference counter */
    }                           /* (there is exactly one facet) */
    facet->edgecnt = i;         /* set the number of edges and */
    assert(chull->edgecnt == i);   /* check it for consistency */
    if (k < 0)                  /* if the edges and the facet normal */
      return 2;                 /* have the correct directions, abort */
    facet->nx = -facet->nx;     /* invert the facet's normal vector */
    facet->ny = -facet->ny;     /* so that it points to the outside */
    facet->nz = -facet->nz;     /* of the tetrahedron to be created */
    for (edge = chull->edges; edge; edge = edge->succ) {
      base            = edge->points[0];
      edge->points[0] = edge->points[1];
      edge->points[1] = base;   /* traverse all edges of the facet */
      edge->dx = -edge->dx;     /* and exchange the end points of */
      edge->dy = -edge->dy;     /* the edge and invert its direction */
      edge->dz = -edge->dz;     /* vector, so that the direction is */
    }                           /* counterclockwise w.r.t. the facet */
    return 2;                   /* abort the function indicating that */
  }                             /* the convex hull needs 3 dimensions */

  /* --- find first visible edge --- */
  for (edge = chull->edges; edge; edge = edge->succ) {
    if (visib2d(facet, edge, curr))
      break;                    /* traverse the facet's edges */
  }                             /* and find a visible edge */
  /* Note that the above loop must terminate in the second step at */
  /* the latest if this is not the first step of the construction, */
  /* because the two added edges come first in the edge list.      */
  assert(edge);                 /* there must be a visible edge */

  /* --- delete visible edges --- */
  chull->edges->pred = chull->last;   /* make the edge list cyclic */
  chull->last->succ  = chull->edges;  /* successor ptr.: clockwise */
  while (visib2d(facet, edge->pred, curr))
    del_edge(chull,edge->pred); /* delete edges counterclockwise */
  while (visib2d(facet, edge->succ, curr))
    del_edge(chull,edge->succ); /* delete edges clockwise */
  chull->edges = edge->succ;    /* make the edge list linear again */
  chull->edges->pred = NULL;    /* (cut the cycle at the anchor, */
  chull->last  = edge;          /* making the anchor the last elem.), */
  edge->succ   = NULL;          /* so that the new edges can be added */
  assert(chull->edgecnt >= 2);  /* at least two edges must be left */

  /* --- extend the convex hull --- */
  r2c = add_edge(chull, chull->edges->points[1], curr);
  c2l = add_edge(chull, curr, edge->pred->points[0]);
  if (!r2c || !c2l) { cleanup(chull, 0); return -1; }
  del_edge(chull, edge);        /* create edges from/to hor. points */
                                /* and delete the anchor edge */
  chull->next = curr->succ;     /* store the next point and */
  return (chull->next) ? 0 : 1; /* return whether convex hull */
}  /* ch_2dim() */              /* is complete (1) or not (0) */

/*--------------------------------------------------------------------*/

static int ch_sorted (CHULL *chull)
{                               /* --- process the next point using */
                                /*     the sorted points algorithm  */
  int    i;                     /* loop variable */
  POINT3 *curr, *next;          /* current and next unprocessed point */
  POINT3 *base;                 /* base point of a facet */
  EDGE   *edge;                 /* to traverse a list/vector of edges */
  EDGE   *anchor = NULL;        /* anchor edge for triangulation */
  FACET  *visib;                /* facet visible from the curr. point */
  FACET  *facet;                /* buffer for facets */
  EDGE   *p0c, *p1c, *cp0;      /* newly created edges */
  FACET  *new;                  /* newly created facet */
  double x, y, z;               /* vector coordinates */
  int    r;                     /* result of function ch_2dim() */

  assert(chull && chull->next); /* check the function argument */

  /* --- update 2 dimensional convex hull --- */
  if (chull->dim <= 2) {        /* if there are only 2 dimensions, */
    r = ch_2dim(chull);         /* extend the 2-dim. convex hull and */
    if (r < 2) return r;        /* check whether it is still 2-dim. */
    chull->dim++;               /* convex hull is now 3-dimensional */
    for (edge = chull->edges; edge; edge = edge->succ)
      edge->points[0]->edge = edge;
    curr   = chull->next;       /* form a counterclockwise cycle, */
    anchor = chull->edges;      /* get the next unprocessed point, */
  }                             /* and the first edge as an anchor */

  /* --- delete visible facets and determine horizon --- */
  else {                        /* if there are already 3 dimensions */
    curr  = chull->next;        /* get the next unprocessed point */
    visib = chull->visible;     /* and the noted visible facet */
    top_facet(chull, visib);    /* move the visible facet to the top */
    do {                        /* find other visible facets */
      for (i = visib->edgecnt; --i >= 0; ) {
        edge = visib->edges[i]; /* traverse the edges of the facet */
        if (edge->refcnt < 2)   /* but skip those, of which the other */
          continue;             /* incident facet has been deleted */
        if (edge->facets[0] == visib)
          reverse(edge);        /* make sure the 2nd facet is visible */
        facet = edge->facets[0];/* get the other incident facet */
        base  = edge->points[0];/* and a base point of this facet */
        x = curr->x -base->x;   /* check whether the facet is visible */
        y = curr->y -base->y;   /* from the curr. point and if it is, */
        z = curr->z -base->z;   /* add it to the list of vis. facets */
        if (x *facet->nx +y *facet->ny +z *facet->nz > EPSILON)
          top_facet(chull, facet);
        else                    /* if the other facet is invisible, */
          base->edge = anchor = edge; /* note the edge as an anchor */
      }                         /* and form a clockwise cyclic list */
      facet = visib;            /* note the current visible facet, */
      visib = visib->pred;      /* get the next visible facet, */
      del_facet(chull, facet);  /* and delete the current one */
    } while (visib);            /* while there is another vis. facet */
  }  /* if (chull->dim <= 2) .. else .. */
  chull->next = next = curr->succ; /* note the next unproc. point */
  chull->visible = NULL;           /* and clear the visible facet */

  /* --- create new edges and facets --- */
  p0c = cp0 = add_edge(chull, anchor->points[0], curr);
  if (!p0c) { cleanup(chull, 0); return -1; }
  edge = anchor;                /* create the first new edge */
  do {                          /* traverse the horizon edges */
    base = edge->points[1];     /* get a base point for the new facet */
    if (base->edge == anchor)   /* if this is the last horizon edge */
      p1c = cp0;                /* get the first created edge */
    else {                      /* if this is not the last hor. edge */
      p1c = add_edge(chull, curr, base);
      if (!p1c) { cleanup(chull, 0); return -1; }
    }                           /* create a new edge */
    facet = edge->facets[0];    /* get the invisible facet and */
    if (fabs(p0c->dx*facet->nx  /* if a new facet would be coplanar, */
    +        p0c->dy*facet->ny  /* extend the invisible facet */
    +        p0c->dz*facet->nz) <= EPSILON)
      new = ext_facet(chull, facet, edge, p0c, p1c);
    else                        /* otherwise create a new facet */
      new = add_facet(chull, edge, p0c, p1c);
    if (!new) { cleanup(chull, 0); return -1; }
    if ( next                   /* if there is another point, but */
    &&  !chull->visible) {      /* no visible facet is known yet, */
      x = next->x -base->x;     /* compute the difference vector to */
      y = next->y -base->y;     /* a point on the extended/created */
      z = next->z -base->z;     /* facet, e.g. the base point */
      if (x *new->nx +y *new->ny +z *new->nz > 0)
        chull->visible = new;   /* if the extended/created facet is */
    }                           /* visible, note it for the next turn */
    p0c  = p1c;                 /* go to the next created edge */
    edge = base->edge;          /* and   the next horizon edge */
  } while (edge != anchor);     /* while not back at the anchor edge */
  assert(!next                  /* if there is another point, */
      || chull->visible);       /* there must be a visible facet */

  return (chull->next) ? 0 : 1; /* return whether convex hull */
}  /* ch_sorted() */            /* is complete (1) or not (0) */

/*--------------------------------------------------------------------*/

static int ch_cgraph (CHULL *chull)
{                               /* --- process the next point using */
                                /*     the conflict graph algorithm */
  int    i;                     /* loop variable */
  CFLE   *cfle;                 /* to traverse the conflict lists */
  POINT3 *curr, *base;          /* current point, base of a facet */
  POINT3 *pt;                   /* to traverse the conflicting points */
  EDGE   *edge, *anchor = NULL; /* to traverse the (horizon) edges */
  FACET  *visib, *invis;        /* facets incident to horizon edge */
  EDGE   *p0c, *p1c, *cp0;      /* newly created edges */
  FACET  *new;                  /* newly created facet */
  double x, y, z;               /* coordinates of difference vector */

  assert(chull && chull->next); /* check the function argument */

  /* --- find a point with conflicts --- */
  for (curr = chull->next; curr; ) { /* traverse the next points */
    if (curr->confls) break;    /* if there are conflicts, abort */
    pt = curr; curr = pt->succ; /* otherwise the next point is */
    del_point(chull, pt);       /* already inside the convex hull */
  }                             /* and hence it can be deleted */
  if (!curr) {                  /* if there are no points left, */
    chull->next = NULL; return 1; }       /* abort the function */
  chull->next = curr->succ;     /* note the next unprocessed point */

  /* --- determine visible facets and horizon --- */
  for (cfle = curr->confls; cfle; cfle = cfle->pt_succ)
    cfle->facet->visible = 1;   /* mark conflicting facets as visible */
  for (cfle = curr->confls; cfle; cfle = cfle->pt_succ) {
    visib = cfle->facet;        /* traverse the conflict facets again */
    for (i = visib->edgecnt; --i >= 0; ) {
      edge = visib->edges[i];   /* traverse the incident edges */
      if (!edge->facets[1]->visible)
        reverse(edge);          /* make sure the 2nd facet is visible */
      if (!edge->facets[0]->visible) {
        edge->points[0]->edge = edge;
        anchor = edge;          /* if only one facet is visible, */
      }                         /* the edge is on the horizon */
    }                           /* (link the horizon edges to form */
  }                             /* a counterclockwise cyclic list) */

  /* --- create new edges and facets --- */
  p0c = cp0 = add_edge(chull, anchor->points[0], curr);
  if (!p0c) { cleanup(chull, 0); return -1; }
  edge = anchor;                /* create the first new edge */
  do {                          /* traverse the horizon edges */
    assert( edge->facets[1]->visible);  /* check the */
    assert(!edge->facets[0]->visible);  /* visibility */
    base = edge->points[1];     /* get a base point for the new facet */
    if (base->edge == anchor)   /* if this is the last horizon edge */
      p1c = cp0;                /* get the first created edge */
    else {                      /* if this is not the last hor. edge */
      p1c = add_edge(chull, curr, base);
      if (!p1c) { cleanup(chull, 0); return -1; }
    }                           /* create a new edge */
    visib = edge->facets[1];    /* get the facets that are */
    invis = edge->facets[0];    /* incident to the horizon edge */
    if (fabs(p0c->dx*invis->nx  /* if a new facet would be coplanar, */
    +        p0c->dy*invis->ny  /* extend the invisible facet */
    +        p0c->dz*invis->nz) <= EPSILON) {
      new = ext_facet(chull, invis, edge, p0c, p1c);
      if (!new) { cleanup(chull, 0); return -1; } }
    else {                      /* if it would not be coplanar, */
      edge->refcnt--;           /* unlink the visible facet and */
      new = add_facet(chull, edge, p0c, p1c);      /* create a  */
      if (!new) { cleanup(chull, 0); return -1; }  /* new facet */

      /* -- build conflict list of new facet -- */
      for (cfle = visib->confls; cfle; cfle = cfle->fc_succ) {
        cfle->point->mark = 0;  /* unmark all conflicting points */
      }                         /* of the currently visible facet */
      for (cfle = invis->confls; cfle; cfle = cfle->fc_succ) {
        pt = cfle->point;       /* traverse the conflicting points of */
        pt->mark = 1;           /* the invisible facet and mark them */
        assert(pt != curr);     /* check for invisibility */
        x = pt->x -base->x;     /* compute the difference vector */
        y = pt->y -base->y;     /* to a point on the created facet, */
        z = pt->z -base->z;     /* e.g. the base point */
        if (x *new->nx +y *new->ny +z *new->nz <= EPSILON)
          continue;             /* skip points from which the */
                                /* created facet is not visible */
        if (!add_confl(pt, new)) { cleanup(chull, 0); return -1; }
      }                         /* add a conflict for other points */
      for (cfle = visib->confls; cfle; cfle = cfle->fc_succ) {
        pt = cfle->point;       /* traverse the conflicting points */
        if ((pt == curr)        /* of the currently visible facet, */
        ||   pt->mark)          /* but skip the current point */
          continue;             /* and already processed points */
        x = pt->x -base->x;     /* compute the difference vector */
        y = pt->y -base->y;     /* to a point on the visible facet, */
        z = pt->z -base->z;     /* e.g. the base point */
        if (x *new->nx +y *new->ny +z *new->nz <= EPSILON)
          continue;             /* skip all points from which the */
                                /* created facet is not visible */
        if (!add_confl(pt, new)) { cleanup(chull, 0); return -1; }
      }                         /* add a conflict for other points */
    }  /* if (p0c->dx ... else .. */
    p0c  = p1c;                 /* go to the next created edge */
    edge = base->edge;          /* and   the next horizon edge */
  } while (edge != anchor);     /* while not back at the anchor edge */

  /* --- delete conflicting facets --- */
  while (curr->confls) del_facet(chull, curr->confls->facet);

  return (chull->next) ? 0 : 1; /* return whether convex hull */
}  /* ch_cgraph() */            /* is complete (1) or not (0) */

/*--------------------------------------------------------------------*/

int ch_next (CHULL *chull)
{                               /* --- process the next point */
  assert(chull);                /* check the function arguments */
  if (!chull->edges)            /* check whether building a convex */
    return -2;                  /* hull has already been started */
  if (!chull->next)             /* if there is no next point, */
    return  1;                  /* the convex hull is complete */
  switch (chull->alg) {         /* depending on the algorithm */
    case CH_SORTED: return ch_sorted(chull);
    default       : return ch_cgraph(chull);
  }                             /* call a specific function */
}  /* ch_nextcg() */

/*--------------------------------------------------------------------*/

int ch_build (CHULL *chull, int alg)
{                               /* --- build a convex hull */
  int r;                        /* result of ch_start() or ch_next() */

  if (!chull->edges)            /* if convex hull construction has */
    r = ch_start(chull, alg);   /* not been started yet, start it, */
  else                          /* otherwise check whether there */
    r = chull->next ? 0 : 1;    /* is an unprocessed point left */
  while (r == 0)                /* while convex hull is not finished */
    r = ch_next(chull);         /* repeat building steps */
  return (r < 0) ? r : 0;       /* return the error status */
}  /* ch_build() */

/*--------------------------------------------------------------------*/

int ch_save (CHULL *chull, const char *fname)
{                               /* --- save points of convex hull */
  POINT3 *pt;                   /* to traverse the list of points */
  FILE   *out;                  /* output file */

  out = fopen(fname, "w");      /* open the output file */
  if (!out) return -1;          /* and check for success */
  for (pt = chull->points; pt; pt = pt->succ)
    fprintf(out, "%g %g %g\n", pt->x, pt->y, pt->z);
  return fclose(out);           /* return write status */
}  /* ch_save() */

/*--------------------------------------------------------------------*/

int ch_load (CHULL *chull, const char *fname)
{                               /* --- load points of convex hull */
  FILE   *in;                   /* input file */
  POINT3 *pt, *tmp;             /* buffers for the created points */
  int    r, err = 0;            /* result of functions, error flag */
  double x, y, z;               /* buffers for coordinates */

  cleanup(chull, 1);            /* clear the convex hull structure */
  in = fopen(fname, "r");       /* open the input file */
  if (!in) return -2;           /* and check for success */
  while (1) {                   /* read loop for points */
    r = fscanf(in, "%lf%lf%lf", &x, &y, &z);
    if (r != 3) { err = (feof(in)) ? 0 : -3; break; }
    if (!add_point(chull, x, y, z)) {
      err = -1; break; }        /* create a point structure */
  }                             /* from the coordinates read */
  fclose(in);                   /* close the input file */
  if (err) { cleanup(chull, 1); return err; }

  for (tmp = pt = chull->points; pt; ) {
    tmp = pt; pt = pt->succ;    /* since the points were added */
    tmp->succ = tmp->pred;      /* to the head of the point list, */
    tmp->pred = pt;             /* they are in reverse order, */
  }                             /* hence the list has to be mirrored */
  chull->points = tmp;          /* set the mirrored list and */
  return 0;                     /* return 'ok' */
}  /* ch_load() */

/*--------------------------------------------------------------------*/

void ch_stats (CHULL *chull)
{                               /* --- print statistics */
  printf("construction algorithm  : %s\n", (chull->alg == CH_SORTED)
         ? "sorted points" : "conflict graph");
  printf("initial number of points: %d\n", chull->init);
  printf("current number of points: %d\n", chull->ptcnt);
  printf("current number of edges : %d\n", chull->edgecnt);
  printf("current number of facets: %d\n", chull->facetcnt);
  printf("number of facets created: %d\n", chull->created);
}  /* ch_stats() */

/*--------------------------------------------------------------------*/
#ifdef CH_DRAW

void ch_draw (CHULL *chull, int visib, PROJ3D *p3d,
              void drawfn (double x, double y, int op))
{                               /* --- draw a convex hull */
  int    i;                     /* loop variable */
  POINT3 *p0, *p1;              /* to traverse the list of points */
  EDGE   *edge, **ep;           /* to traverse the list of edges */
  FACET  *facet;                /* to traverse the list of facets */

  for (p0 = chull->points; p0; p0 = p0->succ)
    p0->mark = p3d_proj(p3d, p0->x, p0->y, p0->z, &p0->px, &p0->py);
                                /* compute projections of all points */
  if (chull->edges) {           /* if there are edges */
    if (chull->facetcnt <= 1)   /* if there is only one facet, */
      visib = CH_WFRAME;        /* always draw a wireframe model */
    if (visib != CH_WFRAME) {   /* if to draw as a solid body */
      for (edge = chull->edges; edge; edge = edge->succ)
        edge->mark = 0;         /* unmark all edges */
      for (facet = chull->facets; facet; facet = facet->succ) {
        p0 = facet->edges[0]->points[0];
        if (p3d_visib(p3d, p0->x,     p0->y,     p0->z,
                           facet->nx, facet->ny, facet->nz) <= 0)
          continue;             /* skip invisible facets */
        for (ep = facet->edges +(i = facet->edgecnt); --i >= 0; )
          (*--ep)->mark = 1;    /* mark all edges */
      }                         /* of a visible facet */
    }                           /* (all visible edges are now marked) */
    for (edge = chull->edges; edge; edge = edge->succ) {
      if ((visib != CH_WFRAME)  /* if to draw as a solid body */
      &&  !edge->mark)          /* and the edge is not marked, */
        continue;               /* skip the edge */
      p0 = edge->points[0]; p1 = edge->points[1];
      if ((p0->mark < 0) || (p1->mark < 0))
        continue;               /* skip edges with marked points */
      drawfn(p0->px, p0->py, 0); drawfn(p1->px, p1->py, 1);
    }                           /* draw a line between the */
  }                             /* end points of each edge */

  p0 = (chull->next || chull->edges) ? chull->next : chull->points;
  for ( ; p0; p0 = p0->succ)    /* mark the unprocessed points */
    if (p0->mark >= 0) drawfn(p0->px, p0->py, 2);
}  /* ch_draw() */

#endif
